@@ -20,6 +20,7 @@ torch.backends.cuda.matmul.allow_tf32 = True
 torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = True
 
 MIN_DALLE_REPO = 'https://huggingface.co/kuprel/min-dalle/resolve/main/'
+IMAGE_TOKEN_COUNT = 256
 
 
 class MinDalle:
@@ -27,13 +28,19 @@ class MinDalle:
         self,
         models_root: str = 'pretrained',
         dtype: torch.dtype = torch.float32,
+        device: str = None,
         is_mega: bool = True, 
         is_reusable: bool = True,
         is_verbose = True
     ):
+        if device == None:
+            device = 'cuda' if torch.cuda.is_available() else 'cpu'
+        if is_verbose: print("using device", device)
+        self.device = device
         self.is_mega = is_mega
         self.is_reusable = is_reusable
         self.dtype = dtype
+        self.vqgan_dtype = torch.float32 if dtype == torch.float16 else dtype
         self.is_verbose = is_verbose
         self.text_token_count = 64
         self.layer_count = 24 if is_mega else 12
@@ -112,12 +119,13 @@ class MinDalle:
             glu_embed_count = self.glu_embed_count,
             text_token_count = self.text_token_count,
             text_vocab_count = self.text_vocab_count,
-            layer_count = self.layer_count
+            layer_count = self.layer_count,
+            device=self.device
         ).to(self.dtype).eval()
         params = torch.load(self.encoder_params_path)
         self.encoder.load_state_dict(params, strict=False)
         del params
-        if torch.cuda.is_available(): self.encoder = self.encoder.cuda()
+        self.encoder = self.encoder.to(device=self.device)
 
 
     def init_decoder(self):
@@ -130,12 +138,12 @@ class MinDalle:
             embed_count = self.embed_count,
             glu_embed_count = self.glu_embed_count,
             layer_count = self.layer_count,
-            start_token = self.image_vocab_count
+            device=self.device
         ).to(self.dtype).eval()
         params = torch.load(self.decoder_params_path)
         self.decoder.load_state_dict(params, strict=False)
         del params
-        if torch.cuda.is_available(): self.decoder = self.decoder.cuda()
+        self.decoder = self.decoder.to(device=self.device)
 
 
     def init_detokenizer(self):
@@ -146,43 +154,37 @@ class MinDalle:
         params = torch.load(self.detoker_params_path)
         self.detokenizer.load_state_dict(params)
         del params
-        if torch.cuda.is_available(): self.detokenizer = self.detokenizer.cuda()
+        self.detokenizer = self.detokenizer.to(device=self.device)
 
 
-    def images_from_tokens(
+    def image_grid_from_tokens(
         self,
         image_tokens: LongTensor,
+        is_seamless: bool,
         is_verbose: bool = False
     ) -> FloatTensor:
         if not self.is_reusable: del self.decoder
-        if torch.cuda.is_available(): torch.cuda.empty_cache()
+        torch.cuda.empty_cache()
         if not self.is_reusable: self.init_detokenizer()
         if is_verbose: print("detokenizing image")
-        images = self.detokenizer.forward(image_tokens).to(torch.uint8)
+        images = self.detokenizer.forward(is_seamless, image_tokens)
         if not self.is_reusable: del self.detokenizer
         return images
 
 
-    def grid_from_images(self, images: FloatTensor) -> Image.Image:
-        grid_size = int(sqrt(images.shape[0]))
-        images = images.reshape([grid_size] * 2 + list(images.shape[1:]))
-        image = images.flatten(1, 2).transpose(0, 1).flatten(1, 2)
-        image = Image.fromarray(image.to('cpu').numpy())
-        return image
-
-
-    def generate_images_stream(
+    def generate_raw_image_stream(
         self, 
         text: str, 
         seed: int,
-        image_count: int,
-        log2_mid_count: int,
+        grid_size: int,
+        progressive_outputs: bool = False,
+        is_seamless: bool = False,
         temperature: float = 1,
         top_k: int = 256,
         supercondition_factor: int = 16,
         is_verbose: bool = False
     ) -> Iterator[FloatTensor]:
-        assert(log2_mid_count in range(5))
+        image_count = grid_size ** 2
         if is_verbose: print("tokenizing text")
         tokens = self.tokenizer.tokenize(text, is_verbose=is_verbose)
         if len(tokens) > self.text_token_count: 
@@ -191,119 +193,97 @@ class MinDalle:
         text_tokens = numpy.ones((2, 64), dtype=numpy.int32)
         text_tokens[0, :2] = [tokens[0], tokens[-1]]
         text_tokens[1, :len(tokens)] = tokens
-
-        text_tokens = torch.tensor(text_tokens).to(torch.long)
-        if torch.cuda.is_available(): text_tokens = text_tokens.cuda()
+        text_tokens = torch.tensor(
+            text_tokens, 
+            dtype=torch.long, 
+            device=self.device
+        )
 
         if not self.is_reusable: self.init_encoder()
         if is_verbose: print("encoding text tokens")
         with torch.cuda.amp.autocast(dtype=self.dtype):
             encoder_state = self.encoder.forward(text_tokens)
         if not self.is_reusable: del self.encoder
-        if torch.cuda.is_available(): torch.cuda.empty_cache()
+        torch.cuda.empty_cache()
 
         if not self.is_reusable: self.init_decoder()
 
         with torch.cuda.amp.autocast(dtype=self.dtype):
-            encoder_state, attention_mask, attention_state, image_tokens = ( 
-                self.decoder.decode_initial(
-                    seed=seed, 
-                    image_count=image_count, 
-                    text_tokens=text_tokens, 
-                    encoder_state=encoder_state
-                )
+            expanded_indices = [0] * image_count + [1] * image_count
+            text_tokens = text_tokens[expanded_indices]
+            encoder_state = encoder_state[expanded_indices]
+            attention_mask = text_tokens.not_equal(1)
+            attention_state = torch.zeros(
+                size=(
+                    self.layer_count,
+                    image_count * 4,
+                    IMAGE_TOKEN_COUNT,
+                    self.embed_count
+                ), 
+                device=self.device
             )
+            image_tokens = torch.full(
+                (IMAGE_TOKEN_COUNT + 1, image_count), 
+                self.image_vocab_count,
+                dtype=torch.long,
+                device=self.device
+            )
+            
+            if seed > 0: torch.manual_seed(seed)
 
-        row_count = 16
-        for row_index in range(row_count):
-            if is_verbose: 
-                print('sampling row {} of {}'.format(row_index + 1, row_count))
+        token_indices = torch.arange(IMAGE_TOKEN_COUNT, device=self.device)
+        settings = torch.tensor(
+            [temperature, top_k, supercondition_factor], 
+            dtype=torch.float32,
+            device=self.device
+        )
+        for i in range(IMAGE_TOKEN_COUNT):                
             with torch.cuda.amp.autocast(dtype=self.dtype):
-                attention_state, image_tokens = self.decoder.decode_row(
-                    row_index,
-                    temperature=temperature,
-                    top_k=top_k,
-                    supercondition_factor=supercondition_factor,
-                    encoder_state=encoder_state,
+                image_tokens[i + 1], attention_state = self.decoder.forward(
+                    settings=settings,
                     attention_mask=attention_mask,
+                    encoder_state=encoder_state,
                     attention_state=attention_state,
-                    image_tokens_sequence=image_tokens
+                    prev_tokens=image_tokens[i],
+                    token_index=token_indices[[i]]
                 )
-            with torch.cuda.amp.autocast(dtype=torch.float32):
-                if ((row_index + 1) * (2 ** log2_mid_count)) % row_count == 0:
-                    tokens = image_tokens[:, 1:]
-                    images = self.images_from_tokens(tokens, is_verbose)
-                    yield images
+
+            with torch.cuda.amp.autocast(dtype=self.vqgan_dtype):
+                if ((i + 1) % 32 == 0 and progressive_outputs) or i + 1 == 256:
+                    yield self.image_grid_from_tokens(
+                        image_tokens=image_tokens[1:].T,
+                        is_seamless=is_seamless,
+                        is_verbose=is_verbose
+                    )
+
+    def generate_image_stream(self, *args, **kwargs) -> Iterator[Image.Image]:
+        image_stream = self.generate_raw_image_stream(*args, **kwargs)
+        for image in image_stream:
+            image = image.to(torch.uint8).to('cpu').numpy()
+            yield Image.fromarray(image)
 
 
-    def generate_image_stream(
-        self, 
-        text: str, 
-        seed: int,
-        grid_size: int,
-        log2_mid_count: int,
-        temperature: float = 1,
-        top_k: int = 256,
-        supercondition_factor: int = 16,
-        is_verbose: bool = False
-    ) -> Iterator[Image.Image]:
-        images_stream = self.generate_images_stream(
-            text=text, 
-            seed=seed,
-            image_count=grid_size ** 2,
-            log2_mid_count=log2_mid_count,
-            temperature=temperature,
-            top_k=top_k,
-            supercondition_factor=supercondition_factor,
-            is_verbose=is_verbose
-        )
-        for images in images_stream:
-            yield self.grid_from_images(images)
+    def generate_images_stream(self, *args, **kwargs) -> Iterator[FloatTensor]:
+        image_stream = self.generate_raw_image_stream(*args, **kwargs)
+        for image in image_stream:
+            grid_size = kwargs['grid_size']
+            image = image.view([grid_size * 256, grid_size, 256, 3])
+            image = image.transpose(1, 0)
+            image = image.reshape([grid_size ** 2, 2 ** 8, 2 ** 8, 3])
+            yield image
 
 
-    def generate_images(
-        self, 
-        text: str,
-        seed: int = -1,
-        image_count: int = 1,
-        temperature: float = 1,
-        top_k: int = 1024,
-        supercondition_factor: int = 16,
-        is_verbose: bool = False
-    ) -> FloatTensor:
-        log2_mid_count = 0
-        images_stream = self.generate_images_stream(
-            text=text,
-            seed=seed,
-            image_count=image_count,
-            temperature=temperature,
-            log2_mid_count=log2_mid_count,
-            top_k=top_k,
-            supercondition_factor=supercondition_factor,
-            is_verbose=is_verbose
-        )
-        return next(images_stream)
-
-
-    def generate_image(
-        self, 
-        text: str,
-        seed: int = -1,
-        grid_size: int = 1,
-        temperature: float = 1,
-        top_k: int = 1024,
-        supercondition_factor: int = 16,
-        is_verbose: bool = False
-    ) -> Image.Image:
-        log2_mid_count = 0
+    def generate_image(self, *args, **kwargs) -> Image.Image:
         image_stream = self.generate_image_stream(
-            text=text,
-            seed=seed,
-            grid_size=grid_size,
-            log2_mid_count=log2_mid_count,
-            temperature=temperature,
-            top_k=top_k,
-            supercondition_factor=supercondition_factor,
-            is_verbose=is_verbose
+            *args, **kwargs, 
+            progressive_outputs=False
         )
-        return next(image_stream)
\ No newline at end of file
+        return next(image_stream)
+
+
+    def generate_images(self, *args, **kwargs) -> Image.Image:
+        images_stream = self.generate_images_stream(
+            *args, **kwargs, 
+            progressive_outputs=False
+        )
+        return next(images_stream)
\ No newline at end of file
